<?php

namespace Drupal\feeds_filemaker\Feeds\Fetcher;

use Drupal\Core\State\StateInterface;
use Drupal\feeds\Exception\EmptyFeedException;
use Drupal\feeds\Feeds\Fetcher\HttpFetcher;
use Drupal\feeds\Result\HttpFetcherResult;
use Drupal\feeds\FeedInterface;
use Drupal\feeds\StateInterface as FeedsStateInterface;
use GuzzleHttp\Exception\RequestException;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use GuzzleHttp\Exception\ServerException;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\key\KeyRepositoryInterface;

/**
 * @FeedsFetcher(
 *   id = "filemaker_api_fetcher",
 *   title = @Translation("FileMaker API Fetcher"),
 *   description = @Translation("Fetches data from FileMaker API with authentication."),
 *   form = {
 *     "configuration" = "Drupal\feeds_filemaker\Feeds\Fetcher\Form\FileMakerApiFetcherForm",
 *     "feed" = "Drupal\feeds\Feeds\Fetcher\Form\HttpFetcherFeedForm",
 *   }
 * )
 */
class FileMakerApiFetcher extends HttpFetcher implements ContainerFactoryPluginInterface {

  protected KeyRepositoryInterface $keyRepository;
  protected StateInterface $coreState;

  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition
  ) {
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->setKeyRepository($container->get('key.repository'));
    $instance->setCoreState($container->get('state'));
    return $instance;
  }

  public function setKeyRepository(KeyRepositoryInterface $key_repository): self {
    $this->keyRepository = $key_repository;
    return $this;
  }

  public function setCoreState(StateInterface $state): self {
    $this->coreState = $state;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function fetch(FeedInterface $feed, FeedsStateInterface $state) {
    $sink = $this->feedsFileSystem->tempnam($feed, 'filemaker_fetcher_');
    $source = $feed->getSource();

    $offset_key = 'filemaker_' . $feed->id() . '_offset';
    $offset = $this->coreState->get($offset_key, 1);
    $initial_offset = $offset;
    $limit = (int) $this->configuration['batch_size'];
    $max_batches = (int) $this->configuration['max_batches'];
    $total_items = 0;
    $batch_count = 0;
    $all_items = [];

    try {
      $token = $this->authenticate();

      do {
        // Determine if we should use GET or POST based on configuration.
        $method = $this->configuration['request_method'] ?? 'GET';
        $options = [
          'headers' => [
            'Authorization' => 'Bearer ' . $token,
          ],
        ];

        // Handle different request methods.
        if ($method === 'POST') {
          // For POST requests, set the content type header.
          $options['headers']['Content-Type'] = 'application/json';

          // Prepare the query as JSON body for POST.
          $query_data = $this->prepareQueryData($this->configuration['query_criteria'] ?? []);
          $query_data['limit'] = $limit;
          $query_data['offset'] = $offset;

          $options['json'] = $query_data;
          $response = $this->client->post($source, $options);
        }
        else {
          // For GET requests, use query parameters.
          $options['query'] = [];

          // Add any additional query parameters from configuration first.
          if (!empty($this->configuration['query_parameters'])) {
            foreach ($this->configuration['query_parameters'] as $param => $value) {
              $options['query'][$param] = $value;
            }
          }

          // Set limit and offset after any custom parameters.
          $options['query']['_limit'] = $limit;
          $options['query']['_offset'] = $offset;

          $response = $this->client->get($source, $options);
        }

        $body = json_decode($response->getBody(), TRUE);

        // Handle different response structures.
        $items = $body['response']['data'] ?? [];

        // For _find endpoint the data is sometimes directly in the response.
        if (empty($items) && isset($body['data'])) {
          $items = $body['data'];
        }

        $item_count = count($items);

        if ($item_count === 0) {
          break;
        }

        $all_items = array_merge($all_items, $items);

        $total_items += $item_count;
        $offset += $item_count;
        $batch_count++;

        // Implement throttling.
        if ($this->configuration['throttle_enabled']) {
          sleep($this->configuration['throttle_delay']);
        }

      } while ($item_count === $limit && ($max_batches === 0 || $batch_count < $max_batches));

      if ($total_items === 0) {
        $this->fileSystem->unlink($sink);
        $state->setMessage($this->t('The feed is empty. Started at offset @start.', ['@start' => $initial_offset]));
        throw new EmptyFeedException();
      }

      // Write all items to the sink file as a single JSON array.
      file_put_contents($sink, json_encode($all_items));

      // Update the offset for the next run.
      if ($item_count === $limit && $max_batches !== 0 && $batch_count === $max_batches) {
        // We've reached the maximum number of batches, but there might be more data.
        $this->coreState->set($offset_key, $offset);
        $state->setMessage($this->t('Imported @count items. Started at offset @start, ended at @end. Reached max batches, next run will start at offset @next.', [
          '@count' => $total_items,
          '@start' => $initial_offset,
          '@end' => $offset - 1,
          '@next' => $offset,
        ]));
      } elseif ($item_count === $limit) {
        // We've reached the item limit but not the max batches (or no max set), so there's likely more data
        $this->coreState->set($offset_key, $offset);
        $state->setMessage($this->t('Imported @count items. Started at offset @start, ended at @end. More data available, next run will start at offset @next.', [
          '@count' => $total_items,
          '@start' => $initial_offset,
          '@end' => $offset - 1,
          '@next' => $offset,
        ]));
      } else {
        // We've reached the end of the data
        $this->coreState->set($offset_key, 1);
        $state->setMessage($this->t('Imported @count items. Started at offset @start, ended at @end. Reached end of data, offset reset to 1.', [
          '@count' => $total_items,
          '@start' => $initial_offset,
          '@end' => $offset - 1,
        ]));
      }

      return new HttpFetcherResult($sink, $response->getHeaders(), $this->fileSystem);
    }
    catch (ServerException $e) {
      $this->fileSystem->unlink($sink);
      if (strpos($e->getMessage(), '"code":"101","message":"Record is missing"') !== FALSE) {
        $this->coreState->set($offset_key, 1);
        if ($total_items > 0) {
          file_put_contents($sink, json_encode($all_items));
          $state->setMessage($this->t('Reached end of data. Imported @count items. Started at offset @start, ended at @end. Offset reset to 1.', [
            '@count' => $total_items,
            '@start' => $initial_offset,
            '@end' => $offset - 1,
          ]));
          return new HttpFetcherResult($sink, [], $this->fileSystem);
        }
        else {
          $state->setMessage($this->t('No items were imported. Started at offset @start. Offset reset to 1.', ['@start' => $initial_offset]));
          throw new EmptyFeedException();
        }
      }
      $state->setMessage($this->t('FileMaker API Server Error: @error. Started at offset @start.', [
        '@error' => $e->getMessage(),
        '@start' => $initial_offset,
      ]));
      throw new EmptyFeedException();
    }
    catch (RequestException $e) {
      $this->fileSystem->unlink($sink);
      $state->setMessage($this->t('The FileMaker API request failed with error: @error. Started at offset @start.', [
        '@error' => $e->getMessage(),
        '@start' => $initial_offset,
      ]));
      throw new EmptyFeedException();
    }
  }

  /**
   * Prepares the query data for POST requests.
   *
   * @param array $criteria
   *   The criteria for the query.
   *
   * @return array
   *   The prepared query data.
   */
  protected function prepareQueryData(array $criteria): array {
    // If criteria is empty, return a default structure
    if (empty($criteria)) {
      return ['query' => []];
    }

    // If criteria is already in the expected format, return it
    if (isset($criteria['query'])) {
      return $criteria;
    }

    // Otherwise, wrap the criteria in a query array
    return ['query' => $criteria];
  }

  protected function authenticate() {
    $auth_endpoint = $this->getCredential('filemaker_auth_endpoint');
    $username = $this->getCredential('filemaker_username');
    $password = $this->getCredential('filemaker_password');

    $response = $this->client->post($auth_endpoint, [
      'headers' => [
        'Authorization' => 'Basic ' . base64_encode("$username:$password"),
      ],
      'json' => [
        'fmDataSource' => [],
      ],
    ]);

    $body = json_decode($response->getBody(), TRUE);
    return $body['response']['token'];
  }

  protected function getCredential(string $key_id): string {
    $key = $this->keyRepository->getKey($key_id);
    if (!$key) {
      throw new \RuntimeException("Key '$key_id' not found.");
    }
    return $key->getKeyValue();
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'batch_size' => 100,
      // 0 means no limit.
      'max_batches' => 0,
      'throttle_enabled' => TRUE,
      'throttle_delay' => 1,
      'request_method' => 'GET',
      'query_criteria' => [],
      'query_parameters' => [],
    ] + parent::defaultConfiguration();
  }
}
