<?php

namespace Drupal\feeds_filemaker\Feeds\Fetcher\Form;

use Drupal\Core\Form\FormStateInterface;
use Drupal\feeds\Feeds\Fetcher\Form\HttpFetcherForm;

class FileMakerApiFetcherForm extends HttpFetcherForm {

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    $form['request_method'] = [
      '#type' => 'select',
      '#title' => $this->t('Request Method'),
      '#description' => $this->t('The HTTP method to use for API requests.'),
      '#options' => [
        'GET' => $this->t('GET'),
        'POST' => $this->t('POST'),
      ],
      '#default_value' => $this->plugin->getConfiguration('request_method') ?? 'GET',
      '#required' => TRUE,
    ];

    $form['query_criteria'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Query Criteria (JSON)'),
      '#description' => $this->t('JSON formatted query criteria for POST requests. Example: {"query":[{"docType":"WEB fiche"}]}'),
      '#default_value' => $this->formatJsonForDisplay($this->plugin->getConfiguration('query_criteria')),
      '#rows' => 5,
      '#states' => [
        'visible' => [
          ':input[name="fetcher_configuration[request_method]"]' => ['value' => 'POST'],
        ],
      ],
    ];

    $form['query_parameters'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Query Parameters'),
      '#description' => $this->t('Additional query parameters for GET requests. One parameter per line in the format "name=value".'),
      '#default_value' => $this->formatQueryParamsForDisplay($this->plugin->getConfiguration('query_parameters')),
      '#rows' => 5,
      '#states' => [
        'visible' => [
          ':input[name="fetcher_configuration[request_method]"]' => ['value' => 'GET'],
        ],
      ],
    ];

    $form['batch_size'] = [
      '#type' => 'number',
      '#title' => $this->t('Batch Size'),
      '#description' => $this->t('The number of items to fetch in each API request.'),
      '#default_value' => $this->plugin->getConfiguration('batch_size'),
      '#min' => 1,
      '#max' => 1000,
      '#required' => TRUE,
    ];

    $form['max_batches'] = [
      '#type' => 'number',
      '#title' => $this->t('Maximum Batches per Run'),
      '#description' => $this->t('The maximum number of batches to process in a single import run. Use 0 for no limit.'),
      '#default_value' => $this->plugin->getConfiguration('max_batches'),
      '#min' => 0,
      '#required' => TRUE,
    ];

    $form['throttle_enabled'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Enable Throttling'),
      '#description' => $this->t('If checked, the import will pause between batches to avoid overwhelming the API.'),
      '#default_value' => $this->plugin->getConfiguration('throttle_enabled'),
    ];

    $form['throttle_delay'] = [
      '#type' => 'number',
      '#title' => $this->t('Throttle Delay'),
      '#description' => $this->t('The number of seconds to wait between batches when throttling is enabled.'),
      '#default_value' => $this->plugin->getConfiguration('throttle_delay'),
      '#min' => 1,
      '#states' => [
        'visible' => [
          ':input[name="fetcher_configuration[throttle_enabled]"]' => ['checked' => TRUE],
        ],
      ],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::validateConfigurationForm($form, $form_state);

    $batch_size = $form_state->getValue('batch_size');
    if ($batch_size < 1 || $batch_size > 1000) {
      $form_state->setErrorByName('batch_size', $this->t('Batch size must be between 1 and 1000.'));
    }

    $max_batches = $form_state->getValue('max_batches');
    if ($max_batches < 0) {
      $form_state->setErrorByName('max_batches', $this->t('Maximum batches must be 0 or a positive integer.'));
    }

    if ($form_state->getValue('throttle_enabled')) {
      $throttle_delay = $form_state->getValue('throttle_delay');
      if ($throttle_delay < 1) {
        $form_state->setErrorByName('throttle_delay', $this->t('Throttle delay must be at least 1 second.'));
      }
    }

    // Validate query criteria JSON for POST requests
    $request_method = $form_state->getValue('request_method');
    if ($request_method === 'POST') {
      $query_criteria = $form_state->getValue('query_criteria');
      if (!empty($query_criteria)) {
        $json = json_decode($query_criteria, TRUE);
        if (json_last_error() !== JSON_ERROR_NONE) {
          $form_state->setErrorByName('query_criteria', $this->t('Query criteria must be valid JSON. Error: @error', [
            '@error' => json_last_error_msg(),
          ]));
        }
        else {
          // Store the decoded JSON to avoid double-encoding
          $form_state->setValue('query_criteria', $json);
        }
      }
    }

    // Validate and parse query parameters for GET requests
    if ($request_method === 'GET') {
      $query_parameters = $form_state->getValue('query_parameters');
      if (!empty($query_parameters)) {
        $params = [];
        $lines = explode("\n", $query_parameters);
        foreach ($lines as $line) {
          $line = trim($line);
          if (empty($line)) {
            continue;
          }

          if (strpos($line, '=') === FALSE) {
            $form_state->setErrorByName('query_parameters', $this->t('Query parameters must be in the format "name=value". Invalid line: @line', [
              '@line' => $line,
            ]));
            break;
          }

          list($name, $value) = explode('=', $line, 2);
          $name = trim($name);
          $value = trim($value);

          if (empty($name)) {
            $form_state->setErrorByName('query_parameters', $this->t('Parameter name cannot be empty. Invalid line: @line', [
              '@line' => $line,
            ]));
            break;
          }

          $params[$name] = $value;
        }

        // Store the parsed parameters
        $form_state->setValue('query_parameters', $params);
      }
    }
  }

  /**
   * Formats JSON data for display in a textarea.
   *
   * @param mixed $data
   *   The data to format.
   *
   * @return string
   *   The formatted JSON string.
   */
  protected function formatJsonForDisplay($data) {
    if (empty($data)) {
      return '';
    }

    if (is_string($data)) {
      // Try to decode and re-encode to format it nicely
      $decoded = json_decode($data, TRUE);
      if (json_last_error() === JSON_ERROR_NONE) {
        return json_encode($decoded, JSON_PRETTY_PRINT);
      }
      return $data;
    }

    return json_encode($data, JSON_PRETTY_PRINT);
  }

  /**
   * Formats query parameters for display in a textarea.
   *
   * @param array $params
   *   The parameters to format.
   *
   * @return string
   *   The formatted parameters string.
   */
  protected function formatQueryParamsForDisplay($params) {
    if (empty($params)) {
      return '';
    }

    if (is_string($params)) {
      return $params;
    }

    $lines = [];
    foreach ($params as $name => $value) {
      $lines[] = "$name=$value";
    }

    return implode("\n", $lines);
  }
}
